#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
Objects complying with swaybar protocol

'''


import typing
import json


class SwayProtoIn():
    '''
    JSON object accepted by sway bar protocol

    Attributes:
        full_text: str: text displayed in segment
        short_text: str: If full_text overflows
        color: str: Text color #RRGGBB[AA]
        background: str: Segment background color #RRGGBB[AA]
        border: str: Segment border color #RRGGBB[AA]
        border_top: int: top border ``px``
        border_bottom: int: bottom border ``px``
        border_left: int: left border ``px``
        border_right: int: right border ``px``
        min_width: int, str: # minimum seg width ``px`` or `THIS WORD SIZE`
        align: str: alignment left (default), right, or center
        name: str: Name to ID segment. For click events
        instance: str: Instance to ID segment. For click events
        urgent: bool: Highlight Seg for urgency
        separator: bool: Visual separation of seg
        separator_block_width: int: padding after seg
        markup: str: pango or none

    Args:
        **kwargs: passed to update method

    Information:
        url: https://www.mankier.com/7/swaybar-protocol

    '''
    def __init__(self, **kwargs) -> None:
        self.full_text: str = ''  # Necessary
        self.short_text: str = None  # If full_text overflows
        self.color: str = None  # Text color #RRGGBB[AA]
        self.background: str = None  # Segment background color #RRGGBB[AA]
        self.border: str = None  # Segment border color #RRGGBB[AA]
        self.border_top: int = 1  # top border <px>
        self.border_bottom: int = 1  # bottom border <px>
        self.border_left: int = 1  # left border <px>
        self.border_right: int = 1  # right border <px>
        # minimum seg width <px> or 'THIS WORD_S SIZE':
        self.min_width: typing.Union[int, str] = None
        self.align: str = 'left'  # left (default), right, or center
        self.name: str = None  # Name to ID segment. For click events
        self.instance: str = None  # Instance to ID segment. For click events
        self.urgent: bool = False  # Highlight Seg for urgency
        self.separator: bool = True  # Visual separation of seg
        self.separator_block_width: int = 9  # padding after seg
        self.markup: str = None  # pango or none
        self.update(**kwargs)

    def update(self, **kwargs) -> None:
        '''
        Update keys from kwargs only if it is an attribute
        else, ignore

        Args:
            full_text: str:  text displayed in segment
            short_text: str: If full_text overflows
            color: str: Text color #RRGGBB[AA]
            background: str: Segment background color #RRGGBB[AA]
            border: str: Segment border color #RRGGBB[AA]
            border_top: int: top border ``px``
            border_bottom: int: bottom border ``px``
            border_left: int: left border ``px``
            border_right: int: right border ``px``
            min_width: int, str:

                * minimum seg width ``px`` or 'THIS WORD_S SIZE':

            align: str: alignment left (default), right, or center
            name: str: Name to ID segment. For click events
            instance: str: Instance to ID segment. For click events
            urgent: bool: Highlight Seg for urgency
            separator: bool: Visual separation of seg
            separator_block_width: int: padding after seg
            markup: str: pango or none

        '''
        self.__dict__ = {key: {**self.__dict__, **kwargs}[key]
                         for key in self.__dict__}

    def __str__(self) -> str:
        '''
        Serialize and Print

        Returns:
            JSON Serial of the object

        '''
        return json.dumps(self.__dict__)


class SwayProtoOut():
    '''
    JSON object delivered by sway bar protocol

    Attributes:
        name: str: The name of seg
        instance: str: The instance of seg
        x: int: The x location that the click occurred at
        y: int: The y location that the click occurred at
        button: int: The x11 button number for the click.
        event: int: The event code of button for the click
        relative_x: int: x wrt segment
        relative_y: int: y wrt segment
        width: int: width of the seg ``px``
        height: int: The height of the seg ``px``
        modifiers: Future form i3bar?

    Args:
        event_json: str: json returned by SwayBar Protocol on mouse-click

    information:
        url: https://www.mankier.com/7/swaybar-protocol

    '''
    def __init__(self, event_json: str) -> None:
        self.name: str = None  # The name of seg
        self.instance: str = None  # The instance of seg
        self.x: int = None  # The x location that the click occurred at
        self.y: int = None  # The y location that the click occurred at
        self.button: int = 0  # The x11 button number for the click.
        self.event: int = None  # The event code of button for the click
        self.relative_x: int = None  # x wrt segment
        self.relative_y: int = None  # y wrt segment
        self.width: int = None  # width of the seg <px>
        self.height: int = None  # The height of the seg <px>
        self.modifiers = None  # Future form i3bar?
        if event_json == '[':
            # stream opened
            return
        if event_json[0] == ',':
            # Continuation from previous infinite stream
            event_json = event_json[1:]
        self.__dict__ = {**self.__dict__, **json.loads(event_json)}
        # NEXT: In python 3.9, this shall simply become
        # self.__dict__ |= json.loads(event_json)



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
Command line parser for pspbar

'''


import typing
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter


def cli() -> typing.Tuple[int, int]:
    '''
    Parse command line

    period: fast update period in seconds
    multi: multiple of <period> that yields slow update period

    Returns:
        tuple(period, multiple)

    '''
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument("period", type=int, default=1, nargs='?',
                        help="period between successive updates in seconds")
    parser.add_argument("-m", "--multi", type=int, default=1,
                        help="multple of 'period' to update other stats")
    args = parser.parse_args()
    return args.period, args.multi



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
CPU monitor

'''


import typing
import psutil
from .classes import BarSeg


class CpuSeg(BarSeg):
    '''
    CPU segment
    '''
    def call_me(self, _=None) -> typing.Dict[str, str]:
        '''
        Create CPU summary string

        Args:
            all are ignored

        Returns:
            dict to update ``BarSeg`` properties

       '''
        return {'magnitude': f"{psutil.cpu_percent():.2f}"}


CPU = CpuSeg(name="cpu", symbol=chr(0x1f9e0), units="%")
'''cpu segment instance'''



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
common shell calls functions
'''


import typing
import subprocess


def notify(info: str, timeout: int = 5,
           send_args: typing.Tuple[str] = ()) -> None:
    '''
    Push ``info`` to notify-send for ``timeout`` seconds

    Args:
        info: str = information to notify
        timeout: int = remove notification after seconds. [0 => permament]
        send_args: arguments passed to notify-send command

    Returns:
        None

    '''
    timeout *= 1000  # miliseconds
    cmd = ['notify-send']
    cmd.extend(send_args)
    if timeout > 0:
        cmd += ['-t', f'{timeout}']
    cmd.append(info)
    subprocess.Popen(
        cmd,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )


def process_comm(*cmd: str,
                 p_name: str = 'processing',
                 timeout: int = None,
                 fail: bool = True, **kwargs) -> str:
    '''
    Generic process definition and communication

    Args:
        *cmd: list(args) passed to subprocess.Popen as first argument
        p_name: notified as 'Error {p_name}: {stderr}
        timeout: communicatoin timeout. If -1, 'communicate' isn't called
        fail: if False, don't exit on error
        **kwargs: passed on to subprocess.Popen

    Returns:
        ``stdout`` from command's communication
        None if stderr and ``fail`` is ``False``
        ``notify`` message if stderr

    '''
    cmd = list(cmd)
    if timeout is not None and timeout < 0:
        process = subprocess.Popen(cmd, **kwargs)  # DONT: *cmd here
        return None
    process = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
        **kwargs
    )
    stdout, stderr = process.communicate(timeout=timeout)
    if stderr:
        notify(f'Error {p_name}: {stderr}')
        if fail:
            return None
    return stdout



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
Load yaml configuration file(s), temp storage folder

Load {action: '--flag', ...} for all available menus
Determine a root directory for temporary storage and log files
'''


import os
import typing
from pathlib import Path
import yaml
from .sway_api import sway_nag


def get_defaults() -> typing.Tuple[str, dict]:
    '''
    Returns:
        swayroot: Confirmed existing Path for sway files
        config: dictionary of default config fetched from default locations

    get default values
    '''
    swayroot = '.'
    config = {}
    def_loc = [
        [os.environ.get('XDG_CONFIG_HOME', None), 'sway'],  # Good practice
        [os.environ.get('HOME', None), '.config', 'sway', ],  # The same thing
        [os.path.join(os.path.dirname(__file__)), 'config'],  # Shipped default
    ]
    for location in def_loc:
        if location[0] is not None:
            swayroot = Path(location[0]).joinpath(*location[1:])
            if swayroot.is_dir():
                break

    # default config
    for location in def_loc:
        if location[0] is not None:
            config = Path(location[0]).joinpath(*location[1:], 'ppsi.yml')
            if config.exists():
                break
    return swayroot, config


def read_config(custom_conf: str = None,
                swayroot: str = None) -> typing.Tuple[Path, dict]:
    '''
    Read ppsi configuration from supplied yml file or default
    Define swayroot to store log files.

    Args:
        custom_conf: custom path of config file ppsi.yml
        swayroot: custom path of root directory to store sway data

    Returns:
        swayroot, config referenced by ``menu``
    '''
    # default locations
    defroot, defconfig = get_defaults()
    if swayroot is None:
        swayroot = defroot
    if custom_conf is None:
        root_path = Path(swayroot).joinpath("ppsi.yml")
        if root_path.exists():
            custom_conf = root_path
    if custom_conf is None:
        custom_conf = defconfig
    with open(custom_conf, "r") as config_h:
        try:
            config = yaml.safe_load(config_h)
        except (FileNotFoundError, yaml.composer.ComposerError) as err:
            sway_nag(msg=err, error=True)
    return Path(swayroot), config



#!/usr/bin/env python3
# -*- coding: utf-8; mode: python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
ppsi client executable

pushes encoded request to
server via unix socket
'''


from . import client_call


if __name__ == '__main__':
    client_call()



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
(WIFI)Internet-monitoring segments
'''


import os
import typing
import psutil
from ..common import shell
from ..server.wificonnect import refresh_wifi
from .classes import BarSeg


NETCHECK = os.path.join(os.path.dirname(__file__), 'shell_dep',
                        'netcheck.sh')


class IpAddrSeg(BarSeg):
    '''
    ip address segment
    '''
    def call_me(self, _=None) -> typing.Dict[str, str]:
        '''
        Create IP ADDRESS string.

        Args:
            all are ignored.

        Returns:
            dictionary to update segment attributes
'''
        color = 0x777777
        stdout = shell.process_comm('bash', NETCHECK, fail=False)
        # print("NET_STATUS:", stdout, stderr)
        if stdout is None:
            return {'vis': False}
        addr = stdout.split("\t")[0]
        net_type = int(stdout.split("\t")[2])
        if net_type & 8:
            # internet connected
            color += 0x007f00  # #007f00
        elif net_type & 4:
            # intranet connected
            color += 0x7f7f00  # #7f7f00
        else:
            color += 0x7f0000  # #7f0000
        if net_type & 2:
            # On home network
            color += 0x00007f  # #00007f  #007f7f  #7f7f7f #7f007f
        elif net_type & 1:
            color += 0x00003f  # #00003f  #007f3f  #7f7f3f #7f003f
        if addr.split(".")[:2] == ["192", "168"]:
            return {'magnitude': ".".join(addr.split(".")[2:]),
                    'color': f"#{hex(color)[2:]}"}
        return {'magnitude': addr}

    def callback(self, _):
        '''
        Update wifi connection, refresh
        '''
        pid = os.fork()
        if pid == 0:
            # child
            refresh_wifi()
            self.update()


class NetSpeedSeg(BarSeg):
    '''
    Net speed segment
    '''
    def call_me(self, mem=None) -> typing.Dict[str, str]:
        '''
        Total internet Speed

        Args:
            mem: memory data held between subsequent calls

        Returns:
            dictionary to update segment attributes
        '''
        net_stats = psutil.net_io_counters()
        if not net_stats:
            return {'mem': mem}
        down_l = net_stats.bytes_recv / 1048576
        up_l = net_stats.bytes_sent / 1048576
        diff = (down_l - mem[1])/mem[0], (up_l - mem[2])/mem[0]
        return {'magnitude': f"{diff[0]:.2f}/{diff[1]:.2f}",
                'mem': [mem[0], down_l, up_l]}


IP_ADDR = IpAddrSeg(name="ip", symbol=chr(0x1f4f6), units="")
'''ip address segment instance'''

NETSPEED = NetSpeedSeg(name="netspeed", symbol='\u21f5\u25BC',
                       units='\u25B2MB/s')
'''netspeed segment instance'''



#!/usr/bin/env python3
# -*- coding: utf-8; mode: python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
test ppsi server

ppsi server
'''


import unittest
from ppsi.common import defined
from ppsi.server.workspaces import (CycleOrder)


class test_server_cycle_order(unittest.TestCase):
    '''
    tests server
    '''
    ws_ord = CycleOrder()

    def test_iadd(self):
        '''
        test method +=
        '''
        # populate
        self.ws_ord += 0
        self.ws_ord += 1
        self.ws_ord += 2
        self.ws_ord += 3
        self.assertEqual(self.ws_ord[-1], 3)
        self.ws_ord += 3
        self.assertEqual(self.ws_ord[-1], 3)
        self.assertEqual(len(self.ws_ord), 4)
        self.ws_ord += 0
        self.assertEqual(len(self.ws_ord), 4)

    def test_reverse(self):
        '''
        test method reverse
        '''
        # populate
        self.ws_ord += 0
        self.ws_ord += 1
        self.ws_ord += 2
        self.ws_ord += 3
        self.ws_ord.reverse()
        self.assertEqual(self.ws_ord[-1], 3)
        self.assertEqual(self.ws_ord[-2], 0)
        self.assertEqual(self.ws_ord[1], 1)
        self.assertEqual(self.ws_ord[0], 2)

    def test_current(self):
        '''
        test method current
        '''
        self.ws_ord += 0
        self.assertEqual(self.ws_ord.current(), 0)

    def test_add(self):
        '''
        test method +
        '''
        self.ws_ord += 0
        self.ws_ord += 1
        self.ws_ord += 2
        self.ws_ord += 3
        self.assertEqual(self.ws_ord + 0, 0)
        self.assertEqual(self.ws_ord + 0, 1)
        self.assertEqual(self.ws_ord + 0, 2)

    def test_sub(self):
        '''
        test method -
        '''
        self.ws_ord += 0
        self.ws_ord += 1
        self.ws_ord += 2
        self.ws_ord += 3
        self.assertEqual(self.ws_ord - 2, 2)
        self.assertEqual(self.ws_ord - 3, 1)
        self.assertEqual(self.ws_ord - 4, 0)



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
power command mode api using launcher_menus
'''


from ..common import shell


syscmd = ['suspend', 'poweroff', 'reboot', 'reboot --firmware-setup']


def adjust_lights():
    '''
    set light to 50%
    '''
    shell.process_comm('light', '-S', '50',
                       p_name='adjusting light', fail=False)


def system(subcmd: int = 1):
    '''
    kwargs:
        subcmd: int = command code for systemctl verbs [0, 1, 2, 3]
            0: suspend
            1: poweroff
            2: reboot
            3: bios_reboot [if available, request reboot to firmware bios]

    Returns:
        None

    Handle system power requests
    '''
    subcmd %= len(syscmd)
    adjust_lights()
    verb = syscmd[subcmd - 1].split(" ")
    shell.process_comm('systemctl', *verb, fail=False)



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
SwayWM Workspace Manager

Guess the active workspace of SwayWM and bind keys according
to supplied yaml configuration

Workspace-order-cycle for quick cycling/switching.

'''


import typing
import json
from ..common import shell, misc
from . import CONFIG
from .sway_api import sway_assign, sway_nag, sway_ws, sway_query, sway_bind


def get_ws() -> str:
    '''
    Get currently focused workspace name by parsing ``swaymsg``

    On Error:
        ``swaynag`` about unavailability of focused workspace

    Returns:
        Workspace Name as displayed (including workspace number)
    '''
    stdout = sway_query('get_workspaces')
    ws_info: typing.List[dict] = json.loads(stdout)
    for w_space in ws_info:
        if w_space['focused']:
            return w_space['name']
    sway_nag('No workspace appears to be focused', error=True)
    return None


class WorkSpace():
    '''
    Workspace object
    Args:
        index: index (number) of workspace (workspace is bound to this key)
        name: displayed name of wordspace
        suicidal: forget this workspace if unfocused and without containers
        keybindings: "keypress": "binding execution"
        assignments: class|app_id: list(container patterns)
    '''
    def __init__(self, index: str, name: str = None,
                 suicidal: bool = False, keybindings: dict = {},
                 assignments: typing.Dict[str, list] = {}) -> None:
        self.index = index
        self.name = self.index
        if name is not None:
            self.name += f":{name}"
        self.suicidal = suicidal
        self.keybindings = keybindings
        self.assignments = assignments

    def bind(self) -> None:
        '''
        Bind all keybindings for this workspace

        Returns:
            None
        '''
        for key, action in self.keybindings.items():
            sway_bind(key, 'exec', action)

    def unbind(self, manager: 'SwayWsMan') -> None:
        '''
        Unbind this workspace's keybindings.
        Destroy suicidal workspaces
        Recipe to run while leaving this workspace.

        Args:
            manager: parent manager
        '''
        for key, action in self.keybindings.items():
            sway_bind(key, 'nop')
        if self.suicidal:
            if self.name not in sway_query('get_workspaces'):
                manager.destroy_ws(self)

    def switch(self) -> typing.Callable:
        '''
        Switch to this workspace.

        Returns:
            Method to execute while switching away
        '''
        sway_ws(self.name)
        self.bind()
        return self.unbind


class CycleOrder(list):
    '''
    Order of visited workspaces:
    0: Oldest workspace' index
    -1: Current workspace' index
    -2: Latest workspace' index [if exists]
    ... and accordingly.

    Items are indices refering to workspaces in
    manager's ``assigned`` list

    Args:
        Parent: Inherit order from

    '''
    def __init__(self, parent: 'CycleOrder' = None) -> None:
        # Current workspace is at index -1
        super().__init__()
        if parent is not None:
            for w_space in parent:
                self += w_space

    def reverse(self) -> None:
        '''
        Reverse the order of visited workspaces protecting current.
        Oldest [0] workspace index becomes latest [-2].
        '''
        rev_order = reversed(self)
        for w_sp in rev_order:
            self += w_sp
        self += self[0]

    def __iadd__(self, value: int) -> 'CycleOrder':
        '''
        Add a workspace index

        Args:
            value: index of newly arrived workspace

        Returns:
            Updated CycleOrder
        '''
        while value in self:
            self.remove(value)
        self.append(value)
        return self

    def current(self) -> int:
        '''
        Retrieve index of currently focused workspace

        Returns:
            current (active) workspace *as registered* in the cycle
            ``None`` if workspace stack is empty
        '''
        if len(self):
            return self[-1]
        else:
            return None

    def __add__(self, index: int = 0) -> int:
        '''
        Query index of oldest workspace

        Args:
            index: index from beginning of stack (the oldest)
            which is now brought to focus in the cycle
        Returns:
            current focus (``index`` brought to focus)
        '''
        if len(self) > index:
            self.__iadd__(self[index])
        return self.current()

    def __sub__(self, index=2) -> int:
        '''
        Query index of latest Nth workspace

        Args:
            index: index from end of stack (the latest)
            which is brought to focus in the cycle
        Returns:
            current focus (``index`` brought to focus)
        '''
        if len(self) >= index:
            self.__iadd__(self[-index])
        return self.current()


class SwayWsMan():
    '''
    Sway Workspace Manager.

        Args:
            config: workspaces config, a part of ppsi.yml
    '''
    def __init__(self, config: dict = CONFIG):
        self.assigned: typing.List[WorkSpace] = []
        self.init_pre_def(config)
        self.bind_base()
        self.cycle_order = CycleOrder()
        self.cycle_order += self.which_workspace((get_ws()))[0]
        self.unbind = lambda *args, **kwargs: None

    @property
    def unbind(self) -> typing.Callable:
        return self._unbind

    @unbind.deleter
    def unbind(self):
        self._unbind = lambda *args, **kwargs: None

    @unbind.setter
    def unbind(self, method):
        self._unbind = self.exitter(method)

    def exitter(self, method: typing.Callable) -> typing.Callable:
        def wrapper(method=method, manager=self, **kwargs):
            return method(manager=manager, **kwargs)
        return wrapper

    @staticmethod
    def bind_base():
        '''
        Default hard bindings
        bind $mod + variable + Tab to cycling through workspaces
        '''
        sway_bind("$mod+Tab",
                  '"exec --no-startup-id ppsi workspace latest"')
        sway_bind("$mod+Shift+Tab",
                  '"exec --no-startup-id ppsi workspace oldest"')
        sway_bind("$mod+Ctrl+Tab",
                  '"exec --no-startup-id ppsi workspace reverse"')

    def init_pre_def(self, config: dict) -> None:
        '''
        Initialize all workspaces as directed in ``config``.
        Bind workspace switchings:
        $mod+0 through $mod+9, $mod+F1 through $mod+F12

        Args
            config: Configuration to initialize and assign workspaces

        Returns:
            ``None``

        '''
        undefined_ws = [str(num_key) for num_key in range(10)]
        undefined_ws += [f'F{num_key}' for num_key in range(1, 13)]
        primary: str = config.get('key-primary', '$mod+Shift+Return')
        pre_def: typing.List[dict] = config.get('workspaces', [])
        for w_space in pre_def:
            bindings: typing.Dict[str, str] = {}
            assignments: typing.Dict[str, list] = {'class': [], 'app_id': []}

            # Others
            for add_bind in w_space.get('bind'):
                bindings[add_bind['key']] = add_bind['exec']

            # Primary overrides if the default keybinding is assigned to others
            bindings[primary] = w_space.get('primary')
            assign = w_space.get('assignments', {})
            if isinstance(assign, list):
                shell.notify('''
                PPSI ERROR: Assignments: <List>.
                Expected: maps {}
                REPLACE leading '-' by <space>
                ''')
                assign = misc.key2dict(keys=assign, vals='class')

            for app, disp_srv in assign.items():
                if disp_srv and any(disp_srv.lower() in w_pat
                                    for w_pat in ('wayland', 'wl', 'app')):
                    disp = 'app_id'
                else:
                    disp = 'class'
                assignments[disp].append(app)

            for index in w_space.get('index', []):
                undefined_ws.remove(str(index))
                self.create_ws(index=str(index),
                               name=w_space.get('name'),
                               keybindings=bindings,
                               assignments=assignments)
        for undef_idx in undefined_ws:
            self.create_ws(index=str(undef_idx))

    def create_ws(self, **kwargs) -> None:
        '''
        Create a structured workspace

        Args:
            **kwargs: passed to initialize ``WorkSpace`` object

        Returns:
            ``None``
        '''
        w_sp = WorkSpace(**kwargs)
        for disp, app_list in w_sp.assignments.items():
            for app in app_list:
                sway_assign(f'[{disp}="{app}"]', w_sp.name)
        sway_bind(
            f'$mod+{w_sp.index}',
            f'"workspace {w_sp.name}, exec --no-startup-id ppsi workspace"'
        )
        sway_bind(f'$mod+Shift+{w_sp.index}', 'move', 'container',
                  'to', 'workspace', w_sp.name)
        self.assigned.append(w_sp)

    def destroy_ws(self, w_sp: WorkSpace) -> None:
        '''
        Unassign ``w_sp`` from this manager
        CAUTION: Program assignments still remain stuck.
        This may cause appearance of zombies for SwayWsMan.
        Switch to a zombie to regularise it.

        Args:
            w_sp: WorkSpace object to destroy

        Returns:
        ``None``
        '''
        sway_bind(f'$mod+{w_sp.index}', 'nop')
        sway_bind(f'$mod+Shift+{w_sp.index}', 'nop')
        idx = self.assigned.index(w_sp)
        self.assigned.remove(w_sp)
        self.cycle_order.remove(idx)

    def which_workspace(self, name: str) -> typing.Tuple[int, WorkSpace]:
        '''
        Args:
            name: name of workspace

        Returns:
            Corresponding WorkSpace object

        Guess and return WorkSpace from its name
        '''
        if name is not None:
            for ws_idx, w_sp in enumerate(self.assigned):
                if name == w_sp.name:
                    return ws_idx, w_sp
        return None, None


PPSI_WS_MAN = SwayWsMan(config=CONFIG)
'''
SwayWsMan instance in use.
'''


def ws_mod(subcmd: int = 0x00, manager: SwayWsMan = PPSI_WS_MAN) -> None:
    '''
    WorkSpace modification requests called by ppsi client

    Args:
        subcmd: int = subcommand {0,1,2,3}
            0x00: Register a switch of workspace, unbind old, bind new
            0x01: Reverse cycle order
            0x02: Switch to oldest workspace in the cycle
            0x03: Switch to the latest workspace (causes back_and_forth)
        manager: Workspace Manager instance

    Returns:
        ``None``
    '''
    if subcmd == 0x00:
        curr_name = get_ws()
        if curr_name is None:
            # Failed to get workspace name and/or corresponding workspace
            return
        current_id, current_ws = manager.which_workspace(curr_name)
        if current_id is None:
            # create anew
            manager.create_ws(index=curr_name, suicidal=True)
            current_id, current_ws = -1, manager.assigned[-1]
        elif current_id == -1:
            # we've switched to already active workspace,
            # hence, we need to do nothing
            return
        manager.cycle_order += current_id
        # unbind old
        manager.unbind()

        # bind new
        current_ws.bind()
        manager.unbind: typing.Callable = current_ws.unbind
        return
    elif subcmd == 0x01:
        manager.cycle_order.reverse()
        return
    elif subcmd == 0x02:
        current_id = manager.cycle_order + 0
    elif subcmd == 0x03:
        current_id = manager.cycle_order - 2
    # current_ws to-be
    sw_2_ws = manager.assigned[current_id]
    manager.unbind()
    manager.unbind: typing.Callable = sw_2_ws.switch()
    return



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
# inspired by pass's passmenu
'''
password manager pass api

Interactive password manager using `menu`
'''


import os
from pathlib import Path
from time import sleep
import re
from launcher_menus import menu
from ..common import shell


def generate_password(user_idx: str = None) -> None:
    '''
    Args:
        user_idx: user-requested password index

    Returns:
        ``None``

    Generate password interactively and copy it to ``wl-copy``
    '''
    ansi_escape = re.compile(r'\x1B\[[0-?]*[ -/]*[@-~]')
    if user_idx is None:
        user_idx = menu(opts=[], prompt='password for')
    if user_idx is None:
        return
    pass_gen = shell.process_comm('pass', 'generate', user_idx,
                                  p_name='generating password')
    pass_text = pass_gen.split('\n')
    pass_text.remove('')
    pass_plain = ansi_escape.sub('', pass_text[-1])
    shell.process_comm('wl-copy', '-o', '-n', '-f', pass_plain,
                       p_name='grabbing password')


def grab_password() -> str:
    '''
    Select password interactively and copy it to ``wl-copy``

    flush ``wl-paste`` after 45 sec to forget it

    Returns:
        if user entered an unknown password index, return it
        else, ``None``
    '''
    prefix = Path(os.environ.get('PASSWORD_STORE_DIR',
                                 f'{os.environ["HOME"]}/.password-store'))
    password_files = []
    for base_path, branches, leaves in os.walk(prefix):
        for file_name in leaves:
            f_name, f_ext = os.path.splitext(file_name)
            if f_ext == '.gpg':
                f_path = base_path + os.sep + file_name.replace('.gpg', '')
                password_files.append(f_path.replace(str(prefix) + os.sep, ''))
    user_idx = menu(opts=password_files, prompt='password for')
    if user_idx is None:
        return
    if user_idx not in password_files:
        return user_idx
    # Password is known
    pass_text = shell.process_comm('pass', 'show', user_idx,
                                   p_name='reading password db')
    shell.process_comm('wl-copy', '-o', '-n', '-f', pass_text,
                       p_name='grabbing password', timeout=-1)
    sleep(45)  # block for 45 seconds
    shell.process_comm('wl-paste', p_name='throwing password', timeout=-1)
    return


def password(subcmd: int = 0) -> None:
    '''
    Handle password requests

    Args:
        subcmd: actions {0,1,2}
            0: grab password else offer generate (default)
            1: grab password else fail
            2: generate password
    Returns:
        ``None``

    '''
    subcmd %= 2
    user_idx = None
    if subcmd < 2:
        user_idx = grab_password()
        if not user_idx:
            return
    if (
            (not subcmd % 2) and
            (menu(opts=['No', 'Yes'], prompt='Create?') == 'Yes')
    ):
        generate_password(user_idx=user_idx)



#!/usr/bin/env python3
# -*- coding: utf-8; mode: python; -*-
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can RRistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.        See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi. If not, see <https://www.gnu.org/licenses/>.
#
'''
light api to handle screen brightness

calls ``light`` for changes
'''


import typing
import subprocess
import math
from ..common import shell


def bar_val_color(
        value: float,
        pig_order: typing.List[str] = ['AA', 'RR', 'GG', 'BB']
) -> str:
    '''
    Process magnitude to ARGB colors.

    Args:
        value: magnitude value of light [%age]
        pig_order: pigment order

    Returns:
        wob input string val(in %) #background #frame #foreground

    '''
    value /= 100  # light yields %age value
    value = max(value, 0)
    # defaults
    bgcol = {'RR': 0x00, 'GG': 0x00, 'BB': 0x00, 'AA': 0xff}
    color = {'RR': 0xff, 'GG': 0xff, 'BB': 0xff, 'AA': 0xff}
    frcol = {'RR': 0xff, 'GG': 0xff, 'BB': 0xff, 'AA': 0xff}

    if value > 1:
        value = 1
    color['RR'] = round((1 - value) * 0xff)
    color['GG'] = round((value) * 0xff)
    color['BB'] = round(math.sin(math.pi*value) * 0xff)
    frcol = color
    bg_str = ''.join([f'{bgcol[pig]:0{2}x}' for pig in pig_order])
    fr_str = ''.join([f'{frcol[pig]:0{2}x}' for pig in pig_order])
    col_str = ''.join([f'{color[pig]:0{2}x}' for pig in pig_order])
    # outputs
    cent_value = round(value * 100)
    return f"{cent_value} #{bg_str} #{fr_str} #{col_str}"


def light(subcmd: int = 1, change: float = 2) -> None:
    '''
    Call ``light``.

    Args:
        subcmd: action codes {0,1,-1}
            0: show brightness level
            1: light up by <change>%
            -1: light down by <change>%
        change: percentage change requested

    Returns:
         ``None``

    '''
    subcmd %= 3
    cmd = ['light']
    direction = ['-G', '-A', '-U'][subcmd]
    cmd += [direction, f'{change}']
    shell.process_comm(*cmd, p_name="adjusting light", timeout=-1)


def light_feedback(wob: subprocess.Popen):
    '''
    Feedback brightness via ``wob``.

    Args:
        wob: process handle to pipe-in wob input string

    Returns:
        ``None``
    '''
    brightness = float(shell.process_comm('light', "-G"))
    wob_in_str = bar_val_color(value=brightness)
    wob.stdin.write(wob_in_str + "\n")
    wob.stdin.flush()



#!/usr/bin/env python3
# -*- coding: utf-8; mode: python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
Psi Pythonic Sway Interface

Still under testing

Person ⟷ Python ⟷ Sway - Interface
to deal with various faculties
that are not offerred by sway such as:

 * work-space cycling
 * work-space-specific program binding
 * visual feedback for volume and light

Employing:
 * a ppsid server that is contacted by
 * a ppsi client

AND

A swaybar-protocol-compatible 'pspbar'
that can call ppsi functions
'''

import subprocess


from . import server
from . import client
from . import pspbar


def check_installations() -> None:
    '''
    check if dependencies are available
    '''
    server.check_installations()
    pspbar.check_installations()
    client.check_installations()


__all__ = [
    'server',
    'client',
    'pspbar'
]

__version__ = '21.1.12'



#!/usr/bin/env python3
# -*- coding: utf-8; mode: python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
ppsi client

ppsi client minimally interacts with a running ppsid server
by pushing encoded bytes to a unix socket
'''


import typing
import socket
from psprint import print
from ..common import shell, defined
from .command_codes import req2bytes
from .command_line import cli


def check_installations() -> None:
    '''
    check if the following dependencies are available:
        * nothing here yet

    '''
    for proc in []:
        if shell.process_comm('command', '-v', proc, fail=False):
            raise FileNotFoundError(f'{proc} not found')


def client_call() -> None:
    '''
    parse command-line and
    communicate encoded request to unix socket

    Send coded request bytes with kwargs as subcommand bits
    to server socket

    if client was called without arguments,
    show short usage message pointing to -h argument
    '''
    req, kwargs = cli()

    if req is None:
        print('usage: ppsi -h')
        return

    commands: typing.List[bytes] = []
    inst, serial_bytes, serial_len = req2bytes(req, **kwargs)
    if inst is None:
        # bad instruction
        return

    if serial_bytes is not None:
        # pass serialized json object
        commands.append(defined.COMM['ACCEPT'])
        commands.append(serial_len)
        commands.append(serial_bytes)

    commands.append(inst)
    commands.insert(0, defined.COMM['OK'])

    # send command
    response: bytes = defined.COMM['OK']
    try:
        CLIENT = socket.socket(family=socket.AF_UNIX, type=socket.SOCK_STREAM)
        CLIENT.connect(str(defined.SOCK_PATH))
        for cmd in commands:
            if not(response and response == defined.COMM['OK']):
                print(
                    f'SERVER replied {int.from_bytes(response, "big"):0{2}x}',
                    mark='warn'
                )
                break
            CLIENT.send(cmd)
            response = CLIENT.recv(defined.INST_SIZE)
        CLIENT.send(defined.COMM['BYE'])
    except (FileNotFoundError, ConnectionRefusedError):
        # scoket file not found
        print('Confirm that the server is running correctly',
              mark='err')
        return
    except BrokenPipeError:
        if response == defined.COMM['BYE']:
            print("Server Adios", mark='info')
            return
        print('Server Closed unexpectedly')



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
Battery monitor and action segment
'''


import typing
import psutil
from ..common import shell
from .classes import BarSeg


EMOJIS = {
    "bat_100": '\uf240',
    "bat_75":  '\uf240',
    "bat_50":  '\uf242',
    "bat_25":  '\uf243',
    "bat_0":   '\uf244',

}


class BatSeg(BarSeg):
    '''
    Battery segment,
    '''

    def _bat_act(self, conn: bool, fill: float, mem: int) -> int:
        '''
        Emergency Actions.

        Args:
            conn: charger connected?
            fill: battery charge fill percentage
            mem: count memory of warning flash notifications

        Returns:
            updated mem

        Notifies:
            ``notify`` emergency multiple times and suspends if critical

        '''
        if conn:
            if fill > 99 and mem < 5:
                mem += 1
                # Send only 5 notifications
                shell.notify('Battery_charged')
        else:
            mem = 0
            if fill < 20:
                shell.notify('Battery Too Low', timeout=0,
                             send_args=('-u', 'critical'))
            elif fill < 10:
                shell.notify('Battery Too Low Suspending Session...',
                             timeout=0, send_args=('-u', 'critical'))
            elif fill < 5:
                shell.process_comm('systemctl', 'suspend',
                                   timeout=-1, fail=False)
        return mem

    def call_me(self, mem: int = None) -> typing.Dict[str, str]:
        '''
        Create Battery summary string

        Args:
            mem: int = count memory of warning flash notifications

        Returns:
            dict to update ``BarSeg`` properties

        '''
        color = None
        sym_pango = ['', '']
        bat_probe = psutil.sensors_battery()
        if not bat_probe:
            return {'symbol': EMOJIS['bat_0'], 'vis': False}
        bat_fill = bat_probe.percent
        bat_conn = bat_probe.power_plugged
        if bat_conn:
            sym_pango = ['<span foreground="#7fffffff">', '</span>']
        # Action
        mem = self._bat_act(conn=bat_conn, fill=bat_fill, mem=mem)
        # returns
        if bat_fill >= 100:
            sym, val, color = EMOJIS['bat_100'], "100", "#7fffffff"
        elif bat_fill > 75:
            sym, val, color = EMOJIS['bat_75'], f"{bat_fill:.2f}", "#ffff7fff"
        elif bat_fill > 50:
            sym, val, color = EMOJIS['bat_50'], f"{bat_fill:.2f}", "#ffaf7fff"
        elif bat_fill > 25:
            sym, val, color = EMOJIS['bat_25'], f"{bat_fill:.2f}", "#ff7f7fff"
        else:
            sym, val, color = EMOJIS['bat_0'], f"{bat_fill:.2f}", "#ff5f5fff"
        sym = sym_pango[0] + sym + sym_pango[1]
        return {'symbol': sym, 'magnitude': val, 'mem': mem, 'color': color}


BATTERY = BatSeg(name="battery",
                 symbol=EMOJIS['bat_0'],
                 units="%",
                 mem=0,
                 pango=True)
'''
battery segment instance
'''
BATTERY.set_proto(markup='pango')



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
CPU load monitor
'''


import typing
import psutil
from .classes import BarSeg


class LoadSeg(BarSeg):
    '''
    Load Segment
    '''
    def call_me(self, _=None) -> typing.Dict[str, str]:
        '''
        Create CPU load summary string

        Args:
            all are ignored

        Returns:
            dict to update ``BarSeg`` properties

        '''
        color = None
        load_avg = list(map(lambda x: x * 100 / psutil.cpu_count(),
                            psutil.getloadavg()))
        if load_avg[0] > 100:
            color = "#ff5f5fff"
        elif load_avg[0] > 80:
            color = "#ffaf7fff"
        elif load_avg[0] > 60:
            color = "#ffff5fff"
        else:
            return {'vis': False}
        value = "|".join((f'{load:.0f}' for load in load_avg))
        return {'magnitude': value, 'color': color}


LOAD = LoadSeg(name="load", symbol=chr(0x1f3cb), units="%")
'''load monitor segment instance'''



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
API-like calls for swaymsg, swaynag

Swaymsg calls through subprocess module
'''


from ..common import shell
import subprocess


def sway_nag(msg: str, **kwargs) -> str:
    '''
    Display panic message on top of screen using swaynag

    Args:
        msg: message to sway on swaynag bar
        **kwargs: passed on to subprocess.Popen

    Returns:
        swaynag's stdout
        None if time-out

    '''
    if kwargs.get('error'):
        bg_col = 'ff3f1f'
        color = '00c0e0'
    elif kwargs.get('info'):
        bg_col = '000080'
        color = '7fafff'
    else:
        bg_col = '000000'
        color = '7f7f7f'
    sway_report = ['swaynag', '--background', bg_col, '--text', color,
                   '--message', msg]
    try:
        stdout = shell.process_comm(*sway_report, timeout=5, **kwargs)
    except subprocess.TimeoutExpired:
        return None
    return stdout


def sway_call(*swaycmd: str, **kwargs) -> str:
    '''
    Subprocess opened with swaymsg

    Args:
        *swaycmd: str = sway command
        **kwargs: passed on to subprocess.Popen

    Returns:
        swaymsg's stdout

    '''
    swaycmd = ['swaymsg'] + list(swaycmd)
    stdout = shell.process_comm(*swaycmd, **kwargs)
    return stdout


def sway_query(*swaycmd, **kwargs) -> str:
    '''
    Subprocess opened with swaymsg -t

    Args:
        *swaycmd: str = query command
    **kwargs: passed on to subprocess.Popen

    Returns:
        swaymsg's stdout

    '''
    swaycmd = ['-t'] + list(swaycmd)
    return sway_call(*swaycmd, **kwargs)


def sway_bind(*swaycmd, **kwargs) -> str:
    '''
    Keybindings called with subprocess swaymsg bindsym

    Args:
        *swaycmd: str = query command
        **kwargs: passed on to subprocess.Popen

    Returns:
        swaymsg's stdout

    '''
    swaycmd = ['bindsym'] + list(swaycmd)
    return sway_call(*swaycmd, **kwargs)


def sway_ws(*swaycmd, **kwargs) -> str:
    '''
    Workspace calls with subprocess swaymsg workspace

    Args:
        *swaycmd: str = query command
        **kwargs: passed on to subprocess.Popen

    Returns:
        swaymsg's stdout

    '''
    swaycmd = ['workspace'] + list(swaycmd)
    return sway_call(*swaycmd, **kwargs)


def sway_assign(*swaycmd, **kwargs) -> str:
    '''
    App_id assignments to workspaces called with subprocess swaymsg assign

    Args:
        *swaycmd: str = query command
        **kwargs: passed on to subprocess.Popen

    Returns:
        swaymsg's stdout

    '''
    swaycmd = ['assign'] + list(swaycmd)
    return sway_call(*swaycmd, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8; mode: python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
ppsi server

server daemon that responds to ppsi client
'''


from ..common import shell
from .read_config import read_config  # default configs


SWAYROOT, CONFIG = read_config(None, None)


def check_availability():
    '''
    check if following dependencies are available
     * swaymsg: sway
     * nmcli: NetworkManager
     * bluetoothctl: bluez
     * pacmd: pulseaudio
     * pactl: pulseaudio
     * light: light
     * wob: wob
     * pass: password manager
     * wl-copy: wayland copy xclip
     * systemctl: systemctl
    '''
    for proc in [
            'swaymsg',  # sway
            'nmcli',  # NetworkManager
            'bluetoothctl',  # bluez
            'pacmd',  # pulseaudio
            'pactl',  # pulseaudio
            'light',  # light
            'wob',  # wob
            'pass',  # password manager
            'wl-copy'  # wayland copy xclip
            'systemctl',  # systemctl
    ]:
        if shell.process_comm('command', '-v', proc, fail=False):
            raise FileNotFoundError(f'{proc} not found')


def server_call():
    '''
    call to start server
    '''
    from .command_line import cli
    newroot, newconfig, kwargs = cli()
    if newroot is not None:
        globals()['SWAYROOT'] = newroot
    if newconfig is not None:
        globals()['CONFIG'] = newconfig

    from .server import start_srv
    start_srv(**kwargs)


__all__ = [
    'CONFIG',
    'SWAYROOT'
]



#!/usr/bin/env python3
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can RRistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.        See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.        If not, see <https://www.gnu.org/licenses/>.
#
'''
Volume api to interface ``pulseaudio``.

Calls ``pacmd`` for information and ``pactl`` for changes.
'''


import re
import subprocess
import math
from ..common import shell


def bar_val_color(left: float, right: float, muted: bool = False,
                  index: int = 0, pig_order=['AA', 'RR', 'GG', 'BB']) -> str:
    '''
    Process sound volume magnitude to #AARRGGBB colors

    Args:
        left: volume of left channel {percent}
        right: volume of right channel {percent}
        muted: mute state
        index: ignored placeholder
        pig_order: pigment order

    Returns:
        wob input string val(%) #background #frame #foreground
    '''
    if right is None or left is None:
        return ''
    if muted is None:
        muted = False  # assume
    value = left + right
    value /= 200  # pacmd sends %age value
    hyper_vol = False
    value = max(value, 0)
    # defaults
    bgcol = {'RR': 0x00, 'GG': 0x00, 'BB': 0x00, 'AA': 0xff}
    color = {'RR': 0xff, 'GG': 0xff, 'BB': 0xff, 'AA': 0xff}
    frcol = {'RR': 0xff, 'GG': 0xff, 'BB': 0xff, 'AA': 0xff}

    if value > 1:
        value = (value - 1) % 1
        hyper_vol = True
    color['RR'] = round((1 - value) * 0xff)
    color['GG'] = round((value) * 0xff)
    color['BB'] = round(math.sin(math.pi*value) * 0xff)
    if muted:
        for pigment, sat in color.items():
            if pigment != 'AA':
                # skip AA
                frcol[pigment] = 0xff - sat
    else:
        frcol = color
    if hyper_vol:
        color['RR'] |= 0xff
        bgcol, color = color, bgcol
    bg_str = ''.join([f'{bgcol[pig]:0{2}x}' for pig in pig_order])
    fr_str = ''.join([f'{frcol[pig]:0{2}x}' for pig in pig_order])
    col_str = ''.join([f'{color[pig]:0{2}x}' for pig in pig_order])
    # outputs
    cent_value = round(value * 100)
    return f"{cent_value} #{bg_str} #{fr_str} #{col_str}"


def get_sink_defaults() -> dict:
    '''
    fetch pulseaudio volume output channel index,
    left channel volume, right channel volume, mute state,
    from ``pacmd`` output

    Returns:
        Namespace dict of fetched values
    '''
    vol_prop = {
        'index': None,
        'left': None,
        'right': None,
        'muted': None
    }
    active_mark = False
    index_pat = re.compile(r':\W+?(\d+?)\n')
    curr_volume_pat = re.compile(r'.+?(\d+?)%.+?(\d+?)%.+?')
    muted_pat = re.compile(r'\W+?muted:\W+?(\w+)')
    sink_list = shell.process_comm("pacmd", "list-sinks",
                                   p_name='Get pa sink info').split("index")
    for sink_data in sink_list:
        if active_mark:
            index = index_pat.findall(sink_data)[0]
            break
        if sink_data.replace(" ", "")[-1] == "*":
            active_mark = True
    if not active_mark:
        return vol_prop
    vol_prop['index'] = index
    volumes = curr_volume_pat.findall(sink_data)
    muted_grp = muted_pat.findall(sink_data)
    if volumes:
        vol_prop['left'] = int(volumes[0][0])
        vol_prop['right'] = int(volumes[0][1])
    if muted_grp:
        vol_prop['muted'] = muted_grp[0].lower() != 'no'
    return vol_prop


def vol(subcmd: int = 1, change: float = 2) -> None:
    '''
    Args:
        subcmd: int = action codes {0,1,-1}
            0: mute
            1: volume up by <change>%
            -1: volume down <change>%
        change: float = percentage change requested

    Returns:
        ``None``
    '''
    subcmd %= 3
    cmd = ['pactl']
    sink_idx = get_sink_defaults()['index']
    if sink_idx is None:
        return
    if subcmd:
        direction = [None, '+', '-'][subcmd]
        cmd += ['set-sink-volume', f"{direction}{change}%"]
    else:
        cmd += ['set-sink-mute', 'toggle']
    cmd.insert(2, str(sink_idx))
    shell.process_comm(*cmd, p_name="adjusting volume", timeout=-1)


def vol_feedback(wob: subprocess.Popen, **kwargs) -> None:
    '''
    Feedback volume via ``wob``

    Args:
        wob: process handle to pipe-in wob input string

    Returns:
        ``None``
    '''
    sink_state = get_sink_defaults()
    wob_in_str = bar_val_color(**sink_state)
    wob.stdin.write(wob_in_str + "\n")
    wob.stdin.flush()



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
Miscellaneous functions and constants for ppsi/pspbar
'''


import typing
from collections import Iterable


def subdict2dict(*subs: dict) -> dict:
    '''
    Convert a list of dictionaries to single dictionary.

    Args:
        *sub: dict [inside an iterable such as list]

    Returns:
        dict of all ``sub`` compiled together
    '''
    key_val_dict = {}
    for sub in subs:
        for key_item, val_item in sub.items():
            key_val_dict[key_item] = val_item
    return key_val_dict


def key2dict(keys: typing.Iterable,
             vals: typing.Union[object, typing.Iterable] = None,
             pairwise: bool = False) -> dict:
    '''
    Convert List of keys [and of values] to a dictionary.

    Args:
        keys: to be converted to dict
        vals: default value to be imparted to every key
            CAUTION: same 'reference' is passed, changing one will change all
        pairwise:
             ``*keys`` -> ``*vals`` pairwise association
            ``vals`` must be iterable, ``len(keys) == len(vals)``

    Raises:
        ValueError: ``keys`` and ``vals`` length mismatches

    Returns:
        Dict[keys, vals]
    '''
    if isinstance(keys, dict):
        return keys
    if all(isinstance(sub, dict) for sub in keys):
        return subdict2dict(*list(keys))
    keys = list(keys)
    key_val_dict = {}
    if pairwise:
        if not isinstance(keys, Iterable):
            raise ValueError("keys not iterable")
        vals = list(vals)
        if len(keys) != len(vals):
            raise ValueError("lengths of keys and vals don't match")
    else:
        vals = [vals] * len(keys)
    for key_item, val_item in zip(keys, vals):
        print(key_item)
        key_val_dict[key_item] = val_item
    return key_val_dict


def val2dict(vals: typing.Iterable,
             keys: typing.Union[object, typing.Iterable] = None,
             pairwise: bool = False) -> dict:
    '''
    Convert List of values [and of keys] to a dictionary.

    Args:
        vals: to be converted to dict
        keys: default key that refers ``vals``
        pairwise:
            ``*keys`` -> ``*vals`` pairwise association
            ``keys`` must be iterable, ``len(keys) == len(vals)``

    Raises:
        KeyError: ``keys`` and ``vals`` length mismatches

    Returns:
        Dict[keys, vals]
    '''
    if isinstance(keys, dict):
        return keys
    if all(isinstance(sub, dict) for sub in vals):
        return subdict2dict(*list(vals))
    vals = list(vals)
    if not pairwise:
        return {keys: vals}
    if not isinstance(keys, Iterable):
        raise KeyError("keys not iterable")
    keys = list(keys)
    if len(keys) != len(list(vals)):
        raise KeyError("lengths of keys and vals don't match")
    key_val_dict = {}
    for key_item, val_item in zip(keys, vals):
        key_val_dict[key_item] = val_item
    return key_val_dict


def iter2dict(in_iter: typing.Iterable, iter_type: str = 'keys',
              in_pair: typing.Union[object, typing.Iterable] = None,
              pairwise: bool = False) -> dict:
    '''
    Convert a list of iterables [and pairs] to dictionary.

    Args:
        in_iter: to be converted to dict
        iter_type: interpret elements as [keys, vals]
        in_pair: to be paired with in_iter
        pairwise:
            ``*in_iter`` <-> ``*in_pair`` pairwise association
            ``in_pair`` must be iterable
            ``len(in_pair) == len(in_iter)``

    Raises:
        KeyError: <in_pair> and <in_iter> length mismatches if iter_type=keys
        ValueError: <in_pair> and <in_iter> length mismatches if iter_type=vals

    Returns:
        Dict[in_pair, in_iter]
    '''
    if iter_type == 'vals':
        return val2dict(vals=in_iter, keys=in_pair, pairwise=pairwise)
    return key2dict(keys=in_iter, vals=in_pair, pairwise=pairwise)



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
pspbar classes


'''


import typing
import sys
import select
from .sway_proto import SwayProtoIn, SwayProtoOut


class BarSeg():
    '''
    Sway Bar Segment object of SBar

    Args:
        name: str: name-handle of the segment
        symbol: str: displayed on bar
        magnitude: str: the string to display
        units: str: suffixed to magnitude
        mem: Any: initial (placeholder) memory to be added to buffer
        ml_tag: str: pango_tag to wrap around magnitude
        call_me: Callable: function that accepts a memory object returns
                (symbol, Magnitude, mem, ?Visible)
        callback: Callable: function called by click on the segment

    Returns:
        ``None``

    '''
    def __init__(self, **kwargs) -> None:
        self.name = None
        self.vis = True
        self.symbol = ''
        self.ml_tag: typing.List[str] = ['', '']
        self.magnitude = ''
        self.units = ''
        self.mem: typing.Any = None
        self.pango: bool = False
        self.__dict__.update(kwargs)
        self.seg_json = SwayProtoIn(name=self.name,
                                    full_text=self.full_text)

    def call_me(self, _=None) -> typing.Dict[str, str]:
        '''
        Args:
            all are ignored

        Dumb call
        '''
        return {}

    def callback(self, _) -> None:
        '''
        Args:
            all are ignored

        void callback
        '''
        return None

    @property
    def full_text(self):
        '''
        Create full_text from symbol, ml_tag, magnitude, units
        '''
        if self.vis:
            return ' '.join(
                [self.symbol,
                 self.ml_tag[0], self.magnitude, self.ml_tag[1],
                 self.units]
            )
        else:
            return None

    @full_text.deleter
    def full_text(self):
        '''
        delete magnitude
        '''
        self.magnitude = ''
        self.ml_tag = ['', '']

    def set_proto(self, **kwargs) -> None:
        '''
        Set protocol keys

        Args:
            **kwargs: passed on to update function of ``SwayProtoIn``

        Returns:
            ``None``

        '''
        self.seg_json.update(**kwargs)

    def update(self, custom: typing.Callable = None, **kwargs) -> None:
        '''
        Update magnitude and symbol

        Args:
            custom: function that updates the segment
            **kwargs:
                symbol: to update
                magnitude: to update
                mem: to update
                ml_tag: pango tag to decorate
                vis: (bool) visibility
                other kwargs are passed on to ``SwayProtoIn``

        '''
        kwargs = (custom or self.call_me)(self.mem)
        for key in 'symbol', 'magnitude', 'mem', 'ml_tag', 'vis':
            if key in kwargs:
                setattr(self, key, kwargs[key])
                del kwargs[key]
        self.seg_json.update(**kwargs, full_text=self.full_text)
        # print(self.seg_json)


class SBar():
    '''
    Sway bar wrapper class
    Attributes:
        bar_str: final json serialized bar_string that is pushed to swaybar
        bar_segs: segments present in instance
        quick_segments: list of segments that update `quickly`
        slow_segments: list of segments that update `slowly`
        static_segments: list of segments that don't update
    '''
    def __init__(self) -> None:
        self.bar_str = ''
        self.bar_segs = []
        self.quick_segs = []
        self.slow_segs = []
        self.static_segs = []

    def add_segs(self, segment: BarSeg,
                 interval: int = 1, position: int = None) -> None:
        '''
        Add segment to bar

        Args:
            position: Position of segment from edge [end]
            interval: Interval of update.

        Numerical code for interval:
            Beginning: 0,
            frequently: 1,
            intermittently: 2.

        Returns:
            ``None``

        '''
        self.bar_segs.insert(position, segment)

        if interval == 1:
            self.quick_segs.append(segment)
        elif interval == 2:
            self.slow_segs.append(segment)
        else:
            self.static_segs.append(segment)

    def _slow_tick(self) -> None:
        '''
        Update slow-updating segments
        '''
        for seg in self.slow_segs:
            seg.update()

    def _quick_tick(self) -> None:
        '''
        Update fast-updating widgets
        '''
        for seg in self.quick_segs:
            seg.update()

    def _static_vals(self) -> None:
        '''
        Define static segments
        '''
        for seg in self.static_segs:
            seg.update()

    def _update_str(self) -> None:
        '''
        Update bar string
        '''
        self.bar_str = f", [{self.bar_str}]"
        self.bar_str = ", [" + (
            ','.join(
                reversed(
                    [str(seg.seg_json)
                     for seg in filter(lambda x: x.vis, self.bar_segs)]
                )
            )
        ) + "]"

    def loop(self, period: int = 1, multi: int = 1) -> None:
        '''
        The bar-update loop
        Args:
            period: Period between two updates in seconds.
            multi: Multiplier to period that gives slow update period.

        Returns:
            ``None``

        '''
        self._static_vals()
        long_per = 0
        while True:
            self._quick_tick()
            if long_per == 0:
                self._slow_tick()
            self._update_str()
            print(self.bar_str, flush=True)
            long_per = (long_per + period) % multi
            # sleep(period)
            trig, _, _ = select.select([sys.stdin], [], [], period)
            if trig:
                feedback = sys.stdin.readline().rstrip("\n")
                sway_comm = SwayProtoOut(feedback)
                for seg in filter(lambda x: x.name == sway_comm.name,
                                  self.bar_segs):
                    seg.callback(sway_comm)



#!/usr/bin/env python3
# -*- coding: utf-8; mode: python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
decode commands and subcommands received via unix socket
'''


from .workspaces import ws_mod
from .remote import call_remote
from .passmenu import password
from .wificonnect import refresh_wifi
from .btconnect import connect_bluetooth
from .volume import vol
from .light import light
from .powermenu import system


def empty(**kwargs) -> None:
    '''
    Empty placeholder command

    Args:
        all are ignored

    Returns:
        ``None``
    '''
    print('Empty command called')
    return None


CMD = {
    0x10: ws_mod,
    0x20: call_remote,
    0x30: password,
    0x40: refresh_wifi,
    0x50: connect_bluetooth,
    0x60: vol,
    0x70: light,
    0x80: empty,
    0x90: empty,
    0xA0: empty,
    0xB0: empty,
    0xC0: empty,
    0xD0: empty,
    0xE0: empty,
    0xF0: system,
}



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
Wifi connection menu

Refresh available wifi APs and offer to connect to one
If known, connect. If unknown, ask for password

'''


import typing
import re
from launcher_menus import menu
from ..common import shell


def query_known() -> typing.List[str]:
    '''
    Query nmcli for known wifi APs

    Returns:
        Known Wifi AP names
    '''
    known_aps = []
    conn_pat = re.compile(
        r'(.+?) +?\w{8}-\w{4}-\w{4}-\w{4}-\w{12}  wifi +'
    )
    stdout = shell.process_comm('nmcli', 'connection', 'show',
                                p_name='remembering')
    for connection in stdout.split("\n"):
        wifi_conn = conn_pat.findall(connection)
        if wifi_conn:
            known_aps += wifi_conn
    return known_aps


def query_available() -> typing.List[str]:
    '''
    Query nmcli for available wifi APs

    Returns:
        Available Wifi AP names
    '''
    available_aps = []
    info_pat = re.compile(
        r'\*? +(?:\w{2}:){5}\w{2} +(.+?) +\w+? +\d+ +.+? .+? +(\d+?) +'
    )
    stdout = shell.process_comm(
        'nmcli', 'device', 'wifi', 'list', '--rescan', 'yes',
        p_name='discovering'
    )
    for info_str in stdout.split("\n"):
        grab = info_pat.findall(info_str)
        if grab:
            available_aps += grab
    return available_aps


def refresh_wifi(subcmd=None) -> None:
    '''
    Offer a available wifi APs to connect
    if the entered wifi ap is not known, request password

    connection success is flashed via ``notify``

    Args:
        all are ignored

    Returns:
        ``None``

    '''
    known_aps = query_known()
    available_aps = query_available()
    if not available_aps:
        shell.notify('No wifi network available', timeout=0)
        return
    choice = menu(
        opts=[':'.join(info)
              for info in sorted(
                      available_aps,
                      key=(lambda x: int(x[-1])
                           ),
                      reverse=True)],
        prompt="connect to"
    )
    if choice is None:
        return
    choice = choice.split(":")[0]
    cmd = ['nmcli', 'device', 'wifi', 'connect', choice.replace(' ', "\\ ")]
    if choice not in known_aps:
        wifi_pass = menu(
            opts=[],
            prompt="Passsword:",
            background="#000000",
            foreground="#000000"
        )
        cmd += ['password', wifi_pass.replace(' ', "\\ ")]
    stdout = shell.process_comm(*cmd, p_name='connecting')
    if 'error' in stdout.lower():
        shell.notify(f'Error connecting: {stdout}')
    else:
        shell.notify(f'Connected to {choice}')



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
Display date-time clock
'''

import typing
import datetime
from .classes import BarSeg


class TimeSeg(BarSeg):
    '''
    Time Segment

    Attributes:
        full = full 24 hour format?
    '''
    full: bool = True

    def call_me(self, _=None) -> typing.Dict[str, str]:
        '''
        create Time summary string

        Args:
            all are ignored

        Returns:
            dict to update ``BarSeg`` properties

        '''
        if self.full:
            return {'magnitude':
                    datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")}
        return {'magnitude':
                datetime.datetime.now().strftime("%Y-%m-%d %I:%M:%S")}

    def callback(self, feedback) -> None:
        '''
        Toggle 12/24 hour format
        '''
        self.full = not self.full


TIME = TimeSeg(name="time", symbol='\u23f0')
'''time segment instance'''



#!/usr/bin/env python3
# -*- coding: utf-8; mode: python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
command-line parser for ppsi client

parse command line to interpret request and corresponding subcommands
'''


import typing
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter


def cli() -> typing.Tuple[str, dict]:
    '''
    client command line parser
    '''
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    subparsers = parser.add_subparsers()

    _ = subparsers.add_parser(name='help', help='show this help')
    remote = subparsers.add_parser(name='remote', help='connect to remote')
    passwd = subparsers.add_parser(name='pass', help='get passwords')
    wifi = subparsers.add_parser(name='wifi', help='reconnect wifi')
    bluetooth = subparsers.add_parser(name='bluetooth',
                                      help='reconnect bluetooth')
    remote.set_defaults(req='remote')
    passwd.set_defaults(req='pass')
    wifi.set_defaults(req='wifi')
    bluetooth.set_defaults(req='bluetooth')

    workspace = subparsers.add_parser(name='workspace', help='set keybindings')
    workspace.set_defaults(req='workspace')
    workspace.add_argument('mod', type=str, nargs='?', default='update',
                           choices=['update', 'reverse', 'oldest', 'latest'])

    ppsi_comm = subparsers.add_parser(name='comm', help='ppsi instructions')
    ppsi_comm.set_defaults(req='comm')
    ppsi_comm.add_argument('mod', type=str, choices=['reload', 'quit'])

    vol = subparsers.add_parser(name='vol', help='change volume')
    vol.set_defaults(req='vol')
    vol.add_argument('mod', type=str, choices=['up', 'down', 'mute',
                                               '+', '-', '0'])
    vol.add_argument('change', type=int, default=2, nargs="?")

    light = subparsers.add_parser(name='light', help='change light')
    light.set_defaults(req='light')
    light.add_argument('mod', type=str, choices=['up', 'down', 'mute',
                                                 '+', '-', '0'])
    light.add_argument('change', type=int, default=2, nargs='?')

    system = subparsers.add_parser(name='system', help='system utils')
    system.add_argument('mod', choices=['suspend', 'poweroff',
                                        'reboot', 'bios_reboot'])
    system.set_defaults(req='system')
    kwargs = vars(parser.parse_args())
    req: str = kwargs.get('req', 'help')
    if req == 'help':
        parser.print_help()
        req = ''
    if 'req' in kwargs:
        del kwargs['req']
    return req, kwargs



#!/usr/bin/env python3
# -*- coding: utf-8; mode: python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
Global definitions for ppsi socket that are inhereted by both
`ppsid` server and `ppsi` client
'''


import os
from pathlib import Path

# location of socket in a RUNTIME DIRRECTORY
RUN_LOC = [
    os.environ.get('XDG_RUNTIME_DIR'),  # Good practice
    os.path.join('/run/user', os.environ.get('UID', 'current_user')),  # hard
    os.path.join(os.environ.get('HOME'), '.runtime')  # Alas!
]
'''
Default locations of runtime directory:
    * From environment {XDG_RUNTIME_DIR} ELSE
    * as hardcoded /run/user/$UID ELSE
    * as ${HOME}/.runtime
'''

for location in RUN_LOC:
    if location is not None and Path(location).is_dir():
        SOCK_PARENT = Path(location).joinpath(location, 'sway')
        if not SOCK_PARENT.is_dir():
            SOCK_PARENT.mkdir()
        SOCK_PATH: Path = SOCK_PARENT.joinpath('ppsi.sock')
        break

INST_SIZE = 0x1  # 1 byte(s) = 16 communications + 15 commands * 16 subcommands
'''
length (in bytes) of instructions
'''

HEAD_SIZE = 64  # Seriously! this is 2 ** 64 = 16 exabytes!
'''
length (in bytes) of header (that tells the server how many bytes to recv)
'''

CODING = 'utf-8'
'''
character encoding used by server and client
'''

COMM = {
    'OK':     0x00.to_bytes(1, 'big'),  # OK
    'ACCEPT': 0x0A.to_bytes(1, 'big'),  # Please Accept JSON of kwargs
    'BYE':    0x0B.to_bytes(1, 'big'),  # BYE BYE
    'EXIT':   0x0E.to_bytes(1, 'big'),  # Request Exit server
    'FAULT':  0x0F.to_bytes(1, 'big'),  # Fault (ERROR)
}
'''
server-client communication commands 0x00 to 0x0F
'''



#!/usr/bin/env python3
# -*- coding: utf-8; mode: python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
test ppsi client

ppsi client
'''


import unittest
import json
from ppsi.common import defined
from ppsi.client.command_codes import (workspaces, comm, blank, vol, light,
                                       system, req2bytes)


class test_client_command_codes(unittest.TestCase):
    '''
    tests client
    '''
    def test_workspaces(self) -> None:
        '''
        test function workspaces
        '''
        subcmd = {
            'reverse': 0x01,
            'oldest': 0x02,
            'latest': 0x03,
            'back': 0x03,
            'default': 0x00,
        }
        for sub, code in subcmd.items():
            byte_code, serial = workspaces(mod=sub)
            self.assertEqual(byte_code, code)
            self.assertEqual(serial, None)

    def test_comm(self) -> None:
        '''
        test function comm
        '''
        subcmd = {
            'quit': 0x0E,
            'exit': 0x0E,
            'reload': 0x0E,
            'default': None,
        }
        for sub, code in subcmd.items():
            byte_code, serial = comm(mod=sub)
            self.assertEqual(byte_code, code)
            self.assertEqual(serial, None)

    def test_blank(self) -> None:
        '''
        test function blank
        '''
        subcmd = {
            'default': None,
        }
        for sub, code in subcmd.items():
            byte_code, serial = blank(mod=sub)
            self.assertEqual(byte_code, code)
            self.assertEqual(serial, None)

    def test_vol(self) -> None:
        '''
        test function volume
        '''
        subcmd = {
            'mute': 0x00,
            '0': 0x00,
            'up': 0x01,
            '+': 0x01,
            'down': 0x02,
            '-': 0x02,
            'default': None,
        }
        test_change = range(0, 100)
        for sub, code in subcmd.items():
            for change in test_change:
                byte_code, serial = vol(mod=sub, change=change)
                self.assertEqual(byte_code, code)
                self.assertEqual(json.loads(serial), {'change': change})

    def test_light(self) -> None:
        '''
        test function light
        '''
        subcmd = {
            'up': 0x01,
            '+': 0x01,
            'down': 0x02,
            '-': 0x02,
        }
        test_change = range(0, 100)
        for sub, code in subcmd.items():
            for change in test_change:
                byte_code, serial = light(mod=sub, change=change)
                self.assertEqual(byte_code, code)
                self.assertEqual(json.loads(serial), {'change': change})

    def test_system(self) -> None:
        '''
        test function light
        '''
        subcmd = {
            'suspend': 0x01,
            'poweroff': 0x02,
            'reboot': 0x03,
            'bios_reboot': 0x04,
        }
        for sub, code in subcmd.items():
            byte_code, serial = system(mod=sub)
            self.assertEqual(byte_code, code)
            self.assertEqual(serial, None)

    def test_req2bytes(self) -> None:
        '''
        test function req2bytes
        '''
        test_inst, test_serial_bytes, test_serial_len_bytes = req2bytes(
            'light', mod='+', change=5
        )
        serial = json.dumps({'change': 5})
        serial_len = len(serial)
        test_serial_len = int(test_serial_len_bytes.decode(defined.CODING))
        test_serial = test_serial_bytes.decode(defined.CODING)
        self.assertEqual(test_inst, 0x71.to_bytes(defined.INST_SIZE, 'big'))
        self.assertEqual(test_serial, serial)
        self.assertEqual(test_serial_len, serial_len)



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
bluez api to handle bluetooth connections

calls bluetoothctl for connections
'''

import typing
import re
from launcher_menus import menu
from ..common import shell


def query_known() -> typing.Dict[str, str]:
    '''
    Query bluetoothctl for known bluetooth devices

    Returns:
        names: MAC
    '''
    known_devs = {}
    dev_pat = re.compile(
        r'Device ?((?:.{2}:){5}.{2}) +?(.*)'
    )
    bt_ctl_out = shell.process_comm('bluetoothctl', 'devices',
                                    p_name='remembering')
    for bt_dev in bt_ctl_out.split("\n"):
        device = dev_pat.findall(bt_dev)
        if device:
            known_devs[device[0][1]] = device[0][0]
    return known_devs


def connect_bluetooth(subcmd=None) -> None:
    '''
    ``menu`` offers bluetooth devices to connect.
    connection result is flashed via ``notify``.

    Args:
        all are ignored

    Returns:
        ``None``
    '''
    known_devs = query_known()
    choice = menu(opts=known_devs.keys(), prompt="Connect_to:")
    if choice in known_devs.keys():
        # connect to known device
        response = shell.process_comm(
            'bluetoothctl', 'connect', known_devs[choice],
            p_name='connecting'
        )
        if any(k in response.lower() for k in ('fail', 'error')):
            shell.notify(f'Connection to {choice} failed')



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
Call remote commands through ssh/[waypipe]

Interactive username, server, interface, application
that is temporarily remembered till reboot.

TODO: Offer ossify temporary information.

'''


import os
from pathlib import Path
import yaml
import typing
from launcher_menus import menu
from . import CONFIG
from ..common import shell


def handle_tmp_cfg() -> typing.Tuple[str, dict]:
    '''
    look for temporary configuration file to store learnt remote addresses,
    user-names and graphical applications for the session

    Returns:
        location of temporary library
        configuration from temporary library
    '''
    data = {}
    uid = os.environ.get('UID')
    if uid is None:
        # uid couldn't be guessed use temporary directory
        basepath = Path("/tmp")
    else:
        # uid is known
        # try for hard path
        basepath = Path('/run').joinpath('user', str(uid))
    xdg_run: Path = Path(os.environ.get('XDG_RUNTIME_DIR', str(basepath)))
    tmp_cfg = xdg_run.joinpath(
        'sway',
        'ppsi',
        'remote.yml',
    )
    if tmp_cfg.exists():
        with open(tmp_cfg, 'r') as config_h:
            data = yaml.safe_load(config_h)
        if not data:
            data = {}
    return tmp_cfg, data


TMP_CFG, DATA = handle_tmp_cfg()


def add_id(identity: str, new: str = 'user') -> None:
    '''
    Add identity to temporary library.
    This library is maintained in yml format.
    This library gets cleared after system reboot.

    Args:
        identity: new object identity
        new: type of object {'user','host','app'}

    Returns:
        ``None``

    '''
    if not TMP_CFG.exists():
        TMP_CFG.parents[0].mkdir(exist_ok=True)
    if 'remote' not in DATA:
        DATA['remote'] = {'hosts': [], 'users': []}
    if 'apps' not in DATA:
        DATA['apps'] = []
    if new == 'user':
        DATA['remote']['users'].append(identity)
    elif new == 'host':
        DATA['remote']['hosts'].append(identity)
    elif new == 'app':
        DATA['apps'].append(identity)
    with open(TMP_CFG, 'w') as config_h:
        yaml.safe_dump(DATA, config_h)


def which_remote() -> typing.Tuple[str]:
    '''
    Fetch desired remote session details using ``menu``

    Returns:
        username
        host
    '''
    # default config
    known_users: typing.List[str] = CONFIG['remote']['users']
    known_hosts: typing.List[str] = CONFIG['remote']['hosts']

    # temporary config
    if 'remote' in DATA:
        known_users += DATA['remote']['users']
        known_hosts += DATA['remote']['hosts']

    user = menu(opts=known_users, prompt='user') or os.environ['USER']
    host = menu(opts=known_hosts, prompt='host') or 'localhost'
    if user == os.environ['USER'] and host == 'localhost':
        # mind changed
        return None, None
    if user not in known_users:
        add_id(user)
    if host not in known_hosts:
        add_id(host, new='host')
    return user, host


def graphical() -> bool:
    '''
    Get remote interface type: Graphical? Default: False

    Returns:
        Whether graphical interface is demanded [experimental]
    '''
    return menu(opts=['No', 'Yes'], prompt='Graphical') == 'Yes'


def remote_app() -> str:
    '''
    Get remote app, also remember it

    Returns:
        Graphical Remote App
    '''
    known_apps = []
    if 'apps' in DATA:
        known_apps += DATA['apps']
    app = menu(opts=known_apps, prompt='application name')
    if app not in known_apps:
        add_id(app, new='app')
    return app


def call_remote(subcmd=None):
    '''
    Call remote ssh and try to set up a connection
    Use menu launcher to learn
    * username
    * host address
    * interface type {GUI,CLI}
    * Application name (if GUI)

    Args:
        all are ignored

    Returns:
        ``None``

    '''
    user, host = which_remote()
    if host is None or user is None:
        return
    cmd = ['--', 'ssh', f'{user}@{host}']
    if graphical():
        app = remote_app()
        if app is not None:
            cmd.insert(0, 'waypipe')
            cmd.append(
                f'env XDG_SESSION_TYPE=wayland nohup {app} 2>&1 1>/dev/null &'
            )
        else:
            cmd.insert(0, os.environ.get('defterm', 'xterm'))
    else:
        cmd.insert(0, os.environ.get('defterm', 'xterm'))
    shell.process_comm(*cmd, p_name='connecting remote', timeout=-1)



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
Personal Swaybar in Python

Bar that follows swaybar-protocol with an interface to ``ppsi`` functions.

Defines:

  * An interface to input and output data according to swaybar protocol.
  * Simplified Generic segment-object ``BarSeg`` to feed segments.
  * ``SBar`` bar manager to update various segments at different intervals.

'''


import sys
import datetime
import time
import warnings
import json
from ..common import shell
from .command_line import cli
from .classes import SBar, BarSeg
from .battery import BATTERY
from .timer import TIME
from .cpu import CPU
from .ram import RAM
from .network import IP_ADDR, NETSPEED
from .temperature import TEMPERATURE
from .uname import OSNAME
from .load_average import LOAD


def check_installations() -> None:
    '''
    Check if the following dependencies are available:
        * nothing here yet

    '''
    for proc in []:
        if shell.process_comm('command', '-v', proc, fail=False):
            raise FileNotFoundError(f'{proc} not found')


def pspbar():
    '''
    Fetch parameters from cli and launch pspbar

    Returns:
        ``None``

    Main Routine
    '''
    period, multi = cli()
    if 'psutil' not in sys.modules:
        while True:
            # basic output
            print('<span foreground=\\"#ff7f7fff\\"> Install psutil',
                  'meanwhile, falling back to basic:\t',
                  datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                  '</span>', flush=True)
            time.sleep(1)
    else:
        warnings.filterwarnings('ignore')
        topbar = SBar()
        NETSPEED.mem = [period * multi, 0, 0]
        topbar.add_segs(segment=TIME, position=0, interval=1)
        topbar.add_segs(segment=BATTERY, position=1, interval=2)
        topbar.add_segs(segment=CPU, position=2, interval=1)
        topbar.add_segs(segment=TEMPERATURE, position=3, interval=2)
        topbar.add_segs(segment=RAM, position=4, interval=1)
        topbar.add_segs(segment=IP_ADDR, position=5, interval=0)
        topbar.add_segs(segment=NETSPEED, position=6, interval=2)
        topbar.add_segs(segment=OSNAME, position=7, interval=0)
        topbar.add_segs(segment=LOAD, position=8, interval=2)
        header = {'version': 1, "click_events": True}
        print(json.dumps(header), "[", "[]", sep="\n")
        topbar.loop(period=period, multi=multi)


__all__ = [
    'pspbar',
    'BarSeg',
    'SBar',
]



#!/usr/bin/env python3
# -*- coding: utf-8; mode: python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
'''
command-line parser for ppsid server

parse command line to
    interpret server configuration, environment and
    start ppsid server
'''


from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
from .read_config import read_config


def cli():
    '''
    server command line parser
    '''
    swayroot, config = None, None
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        '-f', '--conf-file', type=str, help='config yaml file',
        default=None)
    parser.add_argument(
        '-r', '--sway-root', type=str, help='root for sway files',
        default=None
    )
    args = parser.parse_args()
    if args.sway_root is not None or args.conf_file is not None:
        swayroot, config = read_config(custom_conf=args.conf_file,
                                       swayroot=args.sway_root)
    del args.sway_root
    del args.conf_file
    return swayroot, config, vars(args)



#!/usr/bin/env python3
# -*- coding: utf-8; mode: python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
PPSI client pre-processing

process command/sub-command to be communicated by client to socket
'''


import typing
import json
from ..common import defined


def workspaces(mod: str = None, **kwargs) -> typing.Tuple[int, str]:
    '''
    Workspace actions.

    Args:
        mod: mode of workspace cycle {reverse,oldest,latest,back,None}:
            `reverse`: reverse the order of workspace cycle.
            `oldest`: switch to the oldest workspace.
            `latest`: switch to the satest workspace (back-and-forth).
            `back`: =latest.
            ``None``: Update keybindings, manager about workspace switch.
        **kwargs: all are ignored

    Returns:
        tuple(code of mod, ``None``)
    '''
    subcmd = {
        'reverse': 0x01,
        'oldest': 0x02,
        'latest': 0x03,
        'back': 0x03
    }.get(mod, 0x00)
    return subcmd, None


def comm(mod: str = None, **kwargs) -> typing.Tuple[int, str]:
    '''
    Communication instructions to ppsid server
    'reload' is Not working yet.

    Args:
        mod: mode of ppsid call {reload,quit,exit}:
            `reload`: reload ppsid server.
            `quit`: quit ppsid server.
            `exit`: =quit.
        **kwargs: all are ignored

    Returns:
        tuple(code of mod, ``None``)
    '''
    mod = kwargs.get('mod')
    mod_menu = {
        'quit': 0x0E,
        'exit': 0x0E,
        'reload': 0x0E,
    }
    subcmd = mod_menu.get(mod)
    return subcmd, None


def blank(**kwargs) -> typing.Tuple[int, str]:
    '''
    Blank placeholder.

    Args:
        all are ignored

    Returns:
        tuple(0, ``None``)

    '''
    # abort
    return 0x00, None


def vol(mod: str = None, **kwargs) -> typing.Tuple[int, str]:
    '''
    Volume action.

    Args:
        mod: mode of volume command {up,down,mute,+,-,0}:
            `up`: increase volume.
            `down`: decrease volume.
            `mute`: mute speaker output channel.
            `+`: =up.
            `-`: =down.
            `0`: =mute.
        **kwargs: all are json serialized and returned
            change: % change in volume requested

    Returns:
        tuple(code of mod, json(kwargs))

    '''
    mod_menu = {
        'mute': 0x00,
        '0': 0x00,
        'up': 0x01,
        '+': 0x01,
        'down': 0x02,
        '-': 0x02,
    }
    subcmd = mod_menu.get(mod)
    return subcmd, json.dumps(kwargs)


def light(mod: str = None, **kwargs) -> typing.Tuple[int, str]:
    '''
    Brightness actions.

    Args:
        mod: mode of light command {up,down,+,-}:
            `up`: increase brightness.
            `down`: decrease brightness.
            `+`: =up.
            `-`: =down.
        **kwargs: all are json serialized and returned.
            change: % change in brightness requested.

    Returns:
        tuple(code of mod, json(kwargs))

    '''
    mod_menu = {
        'up': 0x01,
        '+': 0x01,
        'down': 0x02,
        '-': 0x02,
    }
    subcmd = mod_menu.get(mod)
    return subcmd, json.dumps(kwargs)


def system(mod: str = None, **kwargs) -> typing.Tuple[int, str]:
    '''
    Systemctl instructions.

    Args:
        mod: mode of system call {suspend,poweroff,reboot,bios_reboot}:
            `suspend`: suspend system.
            `poweroff`: poweroff system.
            `reboot`: reboot system.
            `bios_reboot`: reboot system to open in bios [if supported].
        **kwargs: all are ignored

    Returns:
        typing.Tuple[code of mod, ``None``]

    '''
    mod_menu = {
        'suspend': 0x01,
        'poweroff': 0x02,
        'reboot': 0x03,
        'bios_reboot': 0x04,
    }
    sub_byte = mod_menu.get(mod)
    return sub_byte, None


CMD: typing.Dict[str, int] = {
    "comm": 0x00,
    "workspace": 0x10,
    "remote": 0x20,
    "pass": 0x30,
    "wifi": 0x40,
    "bluetooth": 0x50,
    "vol": 0x60,
    "light": 0x70,
    "system": 0xF0,
}


SUB_ARGS: typing.Dict[int, typing.Callable] = {
    0x00: comm,
    0x10: workspaces,
    0x20: blank,
    0x30: blank,
    0x40: blank,
    0x50: blank,
    0x60: vol,
    0x70: light,
    0x80: blank,
    0x90: blank,
    0xA0: blank,
    0xB0: blank,
    0xC0: blank,
    0xD0: blank,
    0xE0: blank,
    0xF0: system,
}


def req2bytes(req: str, **kwargs) -> typing.Tuple[bytes, bytes, bytes]:
    '''
    Convert request to encoded instruction bytes and
    byte-encoded serialized input data for the instruction.

    Args:
        req: request string from command-line
        **kwargs: passed to subcommand coding function for serialization

    Returns:
        tuple:
            instruction byte (bytes encoded) or ``None``,
            json-serialized ``**kwargs`` (bytes encoded) or ``None``,
            length of serial_bytes (bytes encoded) or ``None``

    '''
    serial_bytes, serial_len = None, None
    base = CMD.get(req)
    if base is None:
        # bad request
        return None, None, None
    sub_byte, serial = SUB_ARGS[base](**kwargs)
    if sub_byte is None:
        # bad sub-argument
        return None, None, None

    inst: bytes = (base + sub_byte).to_bytes(defined.INST_SIZE, 'big')

    if serial is not None:
        serial_bytes: bytes = serial.encode(defined.CODING)
        serial_len: bytes = str(len(serial_bytes)).encode(defined.CODING)
        serial_len += b' ' * (defined.HEAD_SIZE - len(serial_len))

    return inst, serial_bytes, serial_len



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
Temperature monitoring and acting segment
'''


import typing
import psutil
from .classes import BarSeg


EMOJIS = {
    "fire":    chr(0x1f525),
    "temp_100": '\uf2c7',
    "temp_75":  '\uf2c8',
    "temp_50":  '\uf2c9',
    "temp_25":  '\uf2ca',
    "temp_0":   '\uf2cb',
}


class TempSeg(BarSeg):
    '''
    Temperature segment
    '''
    def call_me(self, _=None) -> typing.Dict[str, str]:
        '''
        Create Temperature summary string

        Args:
            all are ignored

        Returns:
            dict to update ``BarSeg`` properties

        '''
        color = None
        heat = psutil.sensors_temperatures()['coretemp'][0].current
        if heat > 80:
            sym, val = EMOJIS['fire'], f"{heat:.0f}"
            color = "#ff5f5fff"
        elif heat > 70:
            sym, val = EMOJIS['temp_100'], f"{heat:.0f}"
            color = "#ffaf7fff"
        elif heat > 60:
            sym, val = EMOJIS['temp_75'], f"{heat:.0f}"
            color = "#ffff5fff"
        elif heat > 50:
            sym, val = EMOJIS['temp_50'], f"{heat:.0f}"
        elif heat > 40:
            sym, val = EMOJIS['temp_25'], f"{heat:.0f}"
        else:
            sym, val = EMOJIS['temp_0'], f"{heat:.0f}"
        return {'symbol': sym, 'magnitude': val, 'color': color}


TEMPERATURE = TempSeg(name="temperature",
                      symbol=EMOJIS['fire'],
                      units='\u2103')
'''temperature segment instance'''



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
Watch, log and callback events.

Server tells this script everything that it did.
Some events will trigger function calls:

  * visual feedbacks:

    - volume calls
    - light calls

  * logs:

    - all client comm bytes

'''


import os
import subprocess
from pathlib import Path
from .workspaces import ws_mod
from .volume import vol_feedback
from .light import light_feedback
from . import SWAYROOT


class EventLogger():
    '''
    Event logger container object
    Args:
        wob: subprocess handle for wob pipe
        root: root in which, a hidden log-file '.ppsi.log' is maintained

    '''
    def __init__(self, wob: subprocess.Popen, root: Path = SWAYROOT) -> None:
        self.logfile = Path(os.path.join(root, '.ppsi.log'))
        self.wob = wob
        if self.logfile.exists():
            # old logs will get written to .ppsi.log.last_session
            self.logfile.replace(
                os.path.join(root, '.ppsi.log.last_session')
            )
        self.log = open(self.logfile, 'a')
        self.feedbacks = {
            0x10: (ws_mod, ()),  # workspace changed
            0x60: (vol_feedback, (self.wob, )),  # sound volume
            0x70: (light_feedback, (self.wob, )),  # Screen brightness
        }

    def event_registry(self, *args, **kwargs) -> None:
        '''
        Register events, [callback].

        Callback:
            workspace calls: update workspace manager
            light: visual feedback
            volume: visual feedback

        Args:
            *args: are logged into log-file in the format
                args: arg1
                args: arg2
                ...
            **kwargs: logged into log-file as
                key1: value1
                key2: value2
                ...

        Returns:
            ``None``


        '''
        cmd = kwargs.get('comm')
        for key, val in kwargs.items():
            print(f'{key}: {val:x}', flush=True, file=self.log)
        for val in args:
            print(f'args: {val}', flush=True, file=self.log)
        for code, feedback in self.feedbacks.items():
            if (cmd & 0xF0 == code):
                # strip last hex digit
                feedback[0](*feedback[1])


def open_pipe() -> subprocess.Popen:
    '''
    Open pipe for wob
    '''
    cmd: list = ['wob']
    wobproc = subprocess.Popen(cmd, stdin=subprocess.PIPE, text=True)
    return wobproc



#!/usr/bin/env python3
# -*- coding: utf-8; mode: python -*-
#
# Copyright 2020-2021 Pradyumna Paranjape
# This file is part of ppsi.
#
# ppsi is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ppsi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ppsi.  If not, see <https://www.gnu.org/licenses/>.
#
'''
socket -> ppsid server communications

ppsi servers functions to listen and
decode request/subcommands from unix socket
'''


import os
import sys
import threading
import socket
import psutil
import json
from psprint import print
from ..common import defined
from .commands import CMD
from .event_watcher import open_pipe, EventLogger


def read_json_bytes(pipe: socket) -> dict:
    '''
    kwargs:
        pipe: socket = socket throught which json object arrives

    Returns:
        dict interpreted from incomming json

    Read stream from pipe with handshakes with client
    '''
    pipe.send(defined.COMM['OK'])  # send me size of json bytes
    in_bytes = pipe.recv(defined.HEAD_SIZE)
    if not in_bytes:
        pass
    stream_len = int(in_bytes.decode(defined.CODING))
    pipe.send(defined.COMM['OK'])  # I am ready to accept stream
    json_bytes = pipe.recv(stream_len)
    pipe.send(defined.COMM['OK'])
    return json.loads(json_bytes.decode(defined.CODING))


def handle_cmd(pipe: socket, logger: EventLogger) -> None:
    '''
    kwargs:
        pipe: socket = socket through which client communications arrive
        logger: EventLogger = container object holding logger objects

    Returns:
        None

    Talks to each client (called by threading module)
    '''
    connected = True
    kwargs = {}
    while connected:
        in_bytes = pipe.recv(defined.INST_SIZE)
        if not in_bytes:
            pass
        comm = int.from_bytes(in_bytes, 'big')
        if comm < 0x10:
            # socket communication
            if in_bytes == defined.COMM['ACCEPT']:
                kwargs.update(read_json_bytes(pipe)),
            elif in_bytes == defined.COMM['BYE']:
                connected = False
            elif in_bytes == defined.COMM['EXIT']:
                print('Terminate Server Manually', mark='act')
                pipe.send(defined.COMM['BYE'])
                connected = False
            elif in_bytes == defined.COMM['FAULT']:
                print("Client sent fault, it shouldn't", mark='warn')
                pipe.send(defined.COMM['FAULT'])
            elif in_bytes == defined.COMM['OK']:
                pipe.send(defined.COMM['OK'])
            else:
                print("Bad communication code. This reflects a fault",
                      mark='warn')
                pipe.send(defined.COMM['FAULT'])
        else:
            # ppsi command
            err = CMD[comm & 0xF0](subcmd=comm & 0x0F, **kwargs)
            logger.event_registry(comm=comm)
            if err:
                pipe.send(defined.COMM['FAULT'])
            else:
                pipe.send(defined.COMM['OK'])
    pipe.close()


def manage_socket():
    '''
    If socket file exists, check if ppsid server is running
    If not running, rename socket as .last_session (overwriting target)

    Returns:
        True: if socket can be made available
        False: If Server is running
    '''
    # check if ppsid another is running
    count_daemons = 0
    for proc in psutil.process_iter():
        try:
            # Check if process name contains the given name string.
            if 'ppsid' in proc.name().lower():
                count_daemons += 1
        except (psutil.NoSuchProcess,
                psutil.AccessDenied,
                psutil.ZombieProcess):
            pass
    if count_daemons > 1:
        # discounting THIS daemon
        print("Server is already running", mark='err')
        return True
    defined.SOCK_PATH.replace(str(defined.SOCK_PATH) + ".last_session")
    return False


def start_srv():
    '''
    Server that listens to unix socket and handles encoded client requests
    Spawns threads of separate client communications
    '''
    server = socket.socket(family=socket.AF_UNIX, type=socket.SOCK_STREAM)
    wob = open_pipe()
    logger = EventLogger(wob=wob)
    try:
        server.bind(str(defined.SOCK_PATH))
    except OSError:
        if manage_socket():
            return
        server.bind(str(defined.SOCK_PATH))
    print(f'{defined.SOCK_PATH}', pref='listening', mark_color='lg')
    try:
        server.listen()
        while not server._closed:
            pipe, _ = server.accept()
            thread = threading.Thread(target=handle_cmd, args=(pipe, logger))
            thread.start()
    except KeyboardInterrupt:
        pass
    logger.wob.kill()
    os.remove(defined.SOCK_PATH)
    sys.exit(0)
